# coding=utf-8
""""""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from abilian.core.entities import Entity
from abilian.core.models.subjects import User
from abilian.testing import BaseTestCase

from .service import ActivityService


class Message1(Entity):
    pass


class ActivityTestCase(BaseTestCase):

    def setUp(self):
        BaseTestCase.setUp(self)
        self.activity_service = ActivityService()
        self.activity_service.start()

    def test(self):
        service = self.activity_service
        user = User(email='test@example.com')
        message = Message1(creator=user, owner=user, name='test message')

        with self.session.begin_nested():
            self.session.add(user)
            self.session.add(message)
        service.log_activity(None, user, "post", message)
        self.session.commit()

        m_id, m_type = message.id, message.entity_type

        entries = service.entries_for_actor(user, 10)
        assert len(entries) == 1

        entry = entries[0]
        assert entry.actor == user
        assert entry.actor_id == user.id
        assert entry.object == message
        assert entry.object_type == m_type
        assert entry._fk_object_id == m_id
        assert entry.object_id == m_id
        assert entry.target is None

        # test entry doesn't reference target if its in deleted state
        self.session.delete(message)
        self.session.flush()
        # send activity after flush: activity should not reference an instance in
        # "deleted state"
        service.log_activity(None, user, "delete", message)
        self.session.flush()

        entries = service.entries_for_actor(user, 10)
        assert len(entries) == 2
        entry = entries[1]
        assert entry.object is None
        assert entry.object_type == m_type
        assert entry._fk_object_id is None
        assert entry.object_id == m_id
        assert entry.target is None



from __future__ import absolute_import, division, print_function, \
    unicode_literals

import glob
import hashlib
import logging
import mimetypes
import os
import re
import subprocess
import threading
import traceback
from abc import ABCMeta, abstractmethod
from base64 import decodestring, encodestring
from typing import List

from magic import Magic
from six import add_metaclass, raise_from, text_type

from abilian.services.image import FIT, resize

from .service import ConversionError
from .util import get_tmp_dir, make_temp_file

try:
    from six.moves.xmlrpc_client import ServerProxy
except ImportError:
    ServerProxy = None

logger = logging.getLogger(__name__)


# Quick check for tests
def has_pdftotext():
    dev_null = open('/dev/null', 'wb')
    try:
        status = subprocess.call("pdftotext", stderr=dev_null)
        return status == 0
    except BaseException:
        return False
    finally:
        dev_null.close()


def has_libreoffice():
    dev_null = open('/dev/null', 'wb')
    try:
        status = subprocess.call(
            ["soffice", "--help"],
            stdout=dev_null,
            stderr=dev_null,
        )
        return status == 0
    except BaseException:
        return False
    finally:
        dev_null.close()


HAS_PDFTOTEXT = has_pdftotext()
HAS_LIBREOFFICE = has_libreoffice()


@add_metaclass(ABCMeta)
class Handler(object):
    """Abstract base class for handlers."""

    accepts_mime_types = []  # type: List[str]
    produces_mime_types = []  # type: List[str]

    def __init__(self, *args, **kwargs):
        self.log = logger.getChild(self.__class__.__name__)

    def init_app(self, app):
        pass

    @property
    def TMP_DIR(self):
        return get_tmp_dir()

    @property
    def mime_sniffer(self):
        return Magic(mime=True)

    @property
    def encoding_sniffer(self):
        return Magic(mime_encoding=True)

    def accept(self, source_mime_type, target_mime_type):
        """Generic matcher based on patterns."""

        match_source = False
        match_target = False

        for pat in self.accepts_mime_types:
            if re.match("^{}$".format(pat), source_mime_type):
                match_source = True
                break

        for pat in self.produces_mime_types:
            if re.match("^%s$" % pat, target_mime_type):
                match_target = True
                break

        return match_source and match_target

    @abstractmethod
    def convert(self, key, **kw):
        pass


class PdfToTextHandler(Handler):
    accepts_mime_types = ['application/pdf', 'application/x-pdf']
    produces_mime_types = ['text/plain']

    def convert(self, blob, **kw):
        with make_temp_file(blob) as in_fn, make_temp_file() as out_fn:
            try:
                subprocess.check_call(['pdftotext', in_fn, out_fn])
            except Exception as e:
                raise raise_from(ConversionError('pdftotext failed'), e)

            converted = open(out_fn, 'rb').read()
            encoding = self.encoding_sniffer.from_file(out_fn)

        if encoding in ("binary", None):
            encoding = "ascii"
        try:
            converted_unicode = text_type(converted, encoding, errors="ignore")
        except BaseException:
            traceback.print_exc()
            converted_unicode = text_type(converted, errors="ignore")

        return converted_unicode


class AbiwordTextHandler(Handler):
    accepts_mime_types = ['application/msword']
    produces_mime_types = ['text/plain']

    def convert(self, blob, **kw):
        tmp_dir = self.TMP_DIR
        cur_dir = os.getcwd()
        with make_temp_file(blob, suffix=".doc") as in_fn,\
                make_temp_file(suffix='.txt') as out_fn:
            try:
                os.chdir(str(tmp_dir))
                subprocess.check_call([
                    'abiword',
                    '--to',
                    os.path.basename(out_fn),
                    os.path.basename(in_fn),
                ])
            except Exception as e:
                raise_from(ConversionError('abiword failed'), e)
            finally:
                os.chdir(cur_dir)

            converted = open(out_fn).read()
            encoding = self.encoding_sniffer.from_file(out_fn)

        if encoding in ("binary", None):
            encoding = "ascii"
        try:
            converted_unicode = text_type(converted, encoding, errors="ignore")
        except BaseException:
            traceback.print_exc()
            converted_unicode = text_type(converted, errors="ignore")

        return converted_unicode


class AbiwordPDFHandler(Handler):
    accepts_mime_types = [
        'application/msword',
        'application/vnd.oasis.opendocument.text',
        'text/rtf',
    ]
    produces_mime_types = ['application/pdf']

    def convert(self, blob, **kw):
        cur_dir = os.getcwd()
        with make_temp_file(blob, suffix=".doc") as in_fn,\
                make_temp_file(suffix='.pdf') as out_fn:
            try:
                os.chdir(bytes(self.TMP_DIR))
                subprocess.check_call([
                    'abiword',
                    '--to',
                    os.path.basename(out_fn),
                    os.path.basename(in_fn),
                ])
            except Exception as e:
                raise_from(ConversionError('abiword failed'), e)
            finally:
                os.chdir(cur_dir)

            converted = open(out_fn).read()
            return converted


class ImageMagickHandler(Handler):
    accepts_mime_types = ['image/.*']
    produces_mime_types = ['application/pdf']

    def convert(self, blob, **kw):
        with make_temp_file(blob) as in_fn, make_temp_file() as out_fn:
            try:
                subprocess.check_call(['convert', in_fn, "pdf:" + out_fn])
                converted = open(out_fn, 'rb').read()
                return converted
            except Exception as e:
                raise_from(ConversionError('convert failed'), e)


class PdfToPpmHandler(Handler):
    accepts_mime_types = ['application/pdf', 'application/x-pdf']
    produces_mime_types = ['image/jpeg']

    def convert(self, blob, size=500):
        """Size is the maximum horizontal size."""
        l = []
        with make_temp_file(blob) as in_fn, make_temp_file() as out_fn:
            try:
                subprocess.check_call(['pdftoppm', '-jpeg', in_fn, out_fn])
                l = sorted(glob.glob("%s-*.jpg" % out_fn))

                converted_images = []
                for fn in l:
                    converted = resize(
                        open(fn, 'rb').read(),
                        size,
                        size,
                        mode=FIT,
                    )
                    converted_images.append(converted)

                return converted_images
            except Exception as e:
                raise_from(ConversionError('pdftoppm failed'), e)
            finally:
                for fn in l:
                    try:
                        os.remove(fn)
                    except OSError:
                        pass


class UnoconvPdfHandler(Handler):
    """Handles conversion from office documents (MS-Office, OOo) to PDF.

    Uses unoconv.
    """

    # TODO: add more if needed.
    accepts_mime_types = [
        'application/vnd.oasis.*',
        'application/msword',
        'application/mspowerpoint',
        'application/vnd.ms-powerpoint',
        'application/vnd.ms-excel',
        'application/vnd.ms-office',
        'application/vnd.ms-word',
        'application/ms-excel',
        'application/vnd.openxmlformats-officedocument.*',
        'text/rtf',
    ]
    produces_mime_types = ['application/pdf']
    run_timeout = 60
    _process = None
    unoconv = 'unoconv'

    def init_app(self, app):
        unoconv = app.config.get('UNOCONV_LOCATION')
        found = False
        execute_ok = False

        if unoconv:
            found = os.path.isfile(unoconv)
            if found:
                # make absolute path: avoid errors when running with different
                # CWD
                unoconv = os.path.abspath(unoconv)
                execute_ok = os.access(unoconv, os.X_OK)
                if not execute_ok:
                    self.log.warning('Not allowed to execute "{}", fallback to '
                                     '"unoconv"'.format(unoconv))
            else:
                self.log.warning('Cannot find "{}", fallback to "unoconv"'
                                 ''.format(unoconv))

        if not unoconv or not found or not execute_ok:
            unoconv = 'unoconv'

        self.unoconv = unoconv

    @property
    def unoconv_version(self):
        # Hack for my Mac, FIXME later
        if os.path.exists(
                "/Applications/LibreOffice.app/Contents/program/python",
        ):
            cmd = [
                '/Applications/LibreOffice.app/Contents/program/python',
                '/usr/local/bin/unoconv',
                '--version',
            ]
        else:
            cmd = [self.unoconv, '--version']

        process = subprocess.Popen(cmd, stdout=subprocess.PIPE)
        out, err = process.communicate()
        return out

    def convert(self, blob, **kw):
        """Convert using unoconv converter."""
        timeout = self.run_timeout
        with make_temp_file(blob) as in_fn, \
                make_temp_file(prefix='tmp-unoconv-', suffix=".pdf") as out_fn:

            # Hack for my Mac, FIXME later
            if os.path.exists(
                    "/Applications/LibreOffice.app/Contents/program/python",
            ):
                cmd = [
                    '/Applications/LibreOffice.app/Contents/program/python',
                    '/usr/local/bin/unoconv',
                    '-f',
                    'pdf',
                    '-o',
                    out_fn,
                    in_fn,
                ]
            else:
                cmd = [self.unoconv, '-f', 'pdf', '-o', out_fn, in_fn]

            def run_uno():
                try:
                    self._process = subprocess.Popen(
                        cmd,
                        close_fds=True,
                        cwd=bytes(self.TMP_DIR),
                    )
                    self._process.communicate()
                except Exception as e:
                    logger.error('run_uno error: %s', bytes(e), exc_info=True)
                    raise_from(ConversionError('unoconv failed'), e)

            run_thread = threading.Thread(target=run_uno)
            run_thread.start()
            run_thread.join(timeout)

            try:
                if run_thread.is_alive():
                    # timeout reached
                    self._process.terminate()
                    if self._process.poll() is not None:
                        try:
                            self._process.kill()
                        except OSError:
                            logger.warning("Failed to kill process {}".format(
                                self._process,
                            ))

                    self._process = None
                    raise ConversionError(
                        "Conversion timeout ({})".format(timeout),
                    )

                converted = open(out_fn).read()
                return converted
            finally:
                self._process = None


class LibreOfficePdfHandler(Handler):
    """Handles conversion from office documents (MS-Office, OOo) to PDF.

    Uses LibreOffice in headless mode.
    """

    # TODO: add more if needed.
    accepts_mime_types = [
        'application/vnd.oasis.*',
        'application/msword',
        'application/mspowerpoint',
        'application/vnd.ms-powerpoint',
        'application/vnd.ms-excel',
        'application/vnd.ms-office',
        'application/ms-excel',
        'application/vnd.openxmlformats-officedocument.*',
        'text/rtf',
    ]
    produces_mime_types = ['application/pdf']
    run_timeout = 60
    _process = None
    soffice = 'soffice'

    def init_app(self, app):
        soffice = app.config.get('SOFFICE_LOCATION')
        found = False
        execute_ok = False

        if soffice:
            # make absolute path: avoid errors when running with different CWD
            soffice = os.path.abspath(soffice)
            found = os.path.isfile(soffice)
            if not found:
                self.log.error("Can't find executable {}".format(soffice))

        elif os.path.isfile("/usr/local/bin/soffice"):
            soffice = "/usr/local/bin/soffice"

        elif os.path.isfile("/usr/bin/soffice"):
            soffice = "/usr/bin/soffice"

        if soffice:
            execute_ok = os.access(soffice, os.X_OK)
            if not execute_ok:
                self.log.warning('Not allowed to execute "{}"'.format(soffice))

        else:
            self.log.error("Can't find LibreOffice executable")
            soffice = None

        self.soffice = soffice

    def convert(self, blob, **kw):
        """Convert using soffice converter."""
        timeout = self.run_timeout
        with make_temp_file(blob) as in_fn:

            cmd = [self.soffice, "--headless", "--convert-to", "pdf", in_fn]

            # # TODO: fix this if needed, or remove if not needed
            # if os.path.exists(
            #         "/Applications/LibreOffice.app/Contents/program/python"):
            #     cmd = [
            #         '/Applications/LibreOffice.app/Contents/program/python',
            #         '/usr/local/bin/unoconv', '-f', 'pdf', '-o', out_fn, in_fn
            #     ]

            def run_soffice():
                try:
                    self._process = subprocess.Popen(
                        cmd,
                        close_fds=True,
                        cwd=bytes(self.TMP_DIR),
                    )
                    self._process.communicate()
                except Exception as e:
                    logger.error('soffice error: %s', bytes(e), exc_info=True)
                    raise_from(ConversionError('soffice conversion failed'), e)

            run_thread = threading.Thread(target=run_soffice)
            run_thread.start()
            run_thread.join(timeout)

            try:
                if run_thread.is_alive():
                    # timeout reached
                    self._process.terminate()
                    if self._process.poll() is not None:
                        try:
                            self._process.kill()
                        except OSError:
                            logger.warning("Failed to kill process {}".format(
                                self._process,
                            ))

                    self._process = None
                    raise ConversionError(
                        "Conversion timeout ({})".format(timeout),
                    )

                out_fn = os.path.splitext(in_fn)[0] + ".pdf"
                converted = open(out_fn, 'rb').read()
                return converted
            finally:
                self._process = None


class CloudoooPdfHandler(Handler):
    """Handles conversion from OOo to PDF.

    Highly inefficient since file are serialized in base64 over HTTP.

    Deactivated because it's so hard to set up.

    FIXME: needs cleanup, or removal.
    """

    accepts_mime_types = [r'application/.*']
    produces_mime_types = ['application/pdf']

    # Hardcoded for now
    SERVER_URL = "http://localhost:8011"

    pivot_format_map = {
        "doc": "odt",
        "docx": "odt",
        "xls": "ods",
        "xlsx": "ods",
        "ppt": "odp",
        "pptx": "odp",
    }

    def convert(self, key):
        in_fn = "data/%s.blob" % key
        in_mime_type = open("data/%s.mime" % key).read()
        file_extension = mimetypes.guess_extension(in_mime_type).strip(".")

        data = encodestring(open(in_fn).read())
        proxy = ServerProxy(self.SERVER_URL, allow_none=True)

        if in_mime_type.startswith("application/vnd.oasis.opendocument"):
            data = proxy.convertFile(data, file_extension, 'pdf')
        else:
            pivot_format = self.pivot_format_map[file_extension]
            data = proxy.convertFile(data, file_extension, pivot_format)
            data = proxy.convertFile(data, pivot_format, 'pdf')

        converted = decodestring(data)
        new_key = hashlib.md5(converted).hexdigest()
        with open("data/%s.blob" % new_key, "wb") as fd:
            fd.write(converted)
        return new_key


class WvwareTextHandler(Handler):
    accepts_mime_types = ['application/msword']
    produces_mime_types = ['text/plain']

    def convert(self, blob, **kw):

        with make_temp_file(blob) as in_fn, make_temp_file() as out_fn:
            try:
                subprocess.check_call(['wvText', in_fn, out_fn])
            except Exception as e:
                raise_from(ConversionError('wxText failed'), e)

            converted = open(out_fn).read()

            encoding = self.encoding_sniffer.from_file(out_fn)
            if encoding in ("binary", None):
                encoding = "ascii"
            try:
                converted_unicode = text_type(
                    converted,
                    encoding,
                    errors="ignore",
                )
            except BaseException:
                traceback.print_exc()
                converted_unicode = text_type(converted, errors="ignore")

            return converted_unicode



# coding=utf-8
"""Blob.

References to files stored in a on-disk repository
"""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import hashlib
import uuid
from typing import Optional

import sqlalchemy as sa
from flask_sqlalchemy import BaseQuery
from six import text_type
from sqlalchemy.event import listens_for
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.schema import Column
from sqlalchemy.types import Integer

from abilian.core.models.base import Model
from abilian.core.sqlalchemy import UUID, JSONDict


class BlobQuery(BaseQuery):
    """Query class for :class:`Blob` objects."""

    def by_uuid(self, uuid):
        # type: (str) -> Optional[Blob]
        """Like `.get()` but by uuid.

        :param uuid: a `string` or an `uuid`.
        """
        try:
            return self.filter_by(uuid=uuid).one()
        except NoResultFound:
            return None


class Blob(Model):
    """Model for storing large file content.

    Files are stored on-disk, named after their uuid. Repository is
    located in instance folder/data/files.
    """

    __tablename__ = "blob"
    query_class = BlobQuery

    id = Column(Integer(), primary_key=True, autoincrement=True)
    uuid = Column(UUID(), unique=True, nullable=False, default=uuid.uuid4)
    meta = Column(JSONDict(), nullable=False, default=dict)

    def __init__(self, value=None, *args, **kwargs):
        super(Blob, self).__init__(*args, **kwargs)
        if self.uuid is None:
            self.uuid = uuid.uuid4()

        if self.meta is None:
            self.meta = dict()

        if value is not None:
            self.value = value

    @property
    def file(self):
        """Return :class:`pathlib.Path` object used for storing value."""
        from abilian.services.repository import session_repository as repository
        return repository.get(self, self.uuid)

    @property
    def size(self):
        """Return size in bytes of value."""
        f = self.file
        return f.stat().st_size if f is not None else 0

    @property
    def value(self):
        """Binary value content."""
        v = self.file
        return v.open('rb').read() if v is not None else v

    @value.setter
    def value(self, value, encoding='utf-8'):
        """Store binary content to applications's repository and update
        `self.meta['md5']`.

        :param:content: string, bytes, or any object with a `read()` method
        :param:encoding: encoding to use when content is Unicode
        """
        from abilian.services.repository import session_repository as repository
        repository.set(self, self.uuid, value)
        self.meta['md5'] = text_type(hashlib.md5(self.value).hexdigest())

        if hasattr(value, 'filename'):
            filename = getattr(value, 'filename')
            if isinstance(filename, bytes):
                filename = filename.decode('utf-8')
            self.meta['filename'] = filename

        if hasattr(value, 'content_type'):
            self.meta['mimetype'] = getattr(value, 'content_type')

    @value.deleter
    def value(self):
        """Remove value from repository."""
        from abilian.services.repository import session_repository as repository
        repository.delete(self, self.uuid)

    @property
    def md5(self):
        """Return md5 from meta, or compute it if absent."""
        md5 = self.meta.get('md5')
        if md5 is None:
            md5 = text_type(hashlib.md5(self.value).hexdigest())

        return md5

    def __bool__(self):
        """A blob is considered falsy if it has no file."""
        return self.file is not None and self.file.exists()

    # Py3k compat
    __nonzero__ = __bool__


@listens_for(sa.orm.Session, 'after_flush')
def _blob_propagate_delete_content(session, flush_context):
    deleted = (obj for obj in session.deleted if isinstance(obj, Blob))
    for blob in deleted:
        del blob.value



# coding=utf-8
from __future__ import absolute_import

# TODO: enable (currently this breaks the tests)
# from sqlalchemy.exc import SAWarning
# import warnings
# warnings.simplefilter('error', SAWarning)



# coding=utf-8
""""""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import os
import sys
from pathlib import Path

import pip
import pkg_resources
from flask import current_app, render_template
from pip.vcs import vcs
from six import text_type

from ..panel import AdminPanel


class SysinfoPanel(AdminPanel):
    id = 'sysinfo'
    label = 'System information'
    icon = 'hdd'

    def get(self):
        uname = os.popen("uname -a").read()
        python_version = sys.version.strip()

        packages = []

        for dist in pkg_resources.working_set:
            package = dict(
                name=dist.project_name,
                key=dist.key,
                version=dist.version
                if dist.has_version() else 'Unknown version',
                vcs=None,
            )

            location = text_type(Path(dist.location).absolute())
            vcs_name = vcs.get_backend_name(location)

            if vcs_name:
                vc = vcs.get_backend(vcs_name)()
                try:
                    url = vc.get_url(location)
                except pip.exceptions.InstallationError:
                    url = 'None'
                try:
                    revision = vc.get_revision(location)
                except pip.exceptions.InstallationError:
                    revision = 'None'
                package['vcs'] = dict(
                    name=vcs_name, url=url, revision=revision)

            packages.append(package)
            packages.sort(key=lambda d: d.get('key'))

        config_values = [
            (k, repr(v)) for k, v in sorted(current_app.config.items())
        ]

        return render_template(
            "admin/sysinfo.html",
            python_version=python_version,
            packages=packages,
            uname=uname,
            config_values=config_values,
        )



# coding=utf-8
""""""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from flask import url_for

from abilian.testing import BaseTestCase


class TestViews(BaseTestCase):

    def get_setup_config(self):
        """Called before creating application class."""
        config = BaseTestCase.get_setup_config(self)
        config.NO_LOGIN = True
        return config

    def setUp(self):
        BaseTestCase.setUp(self)
        self.app.services['security'].start()

    def test_home(self):
        response = self.client.get(url_for("admin.dashboard"))
        assert response.status_code == 200

    def test_sysinfo(self):
        response = self.client.get(url_for("admin.sysinfo"))
        assert response.status_code == 200

    def test_login_session(self):
        response = self.client.get(url_for("admin.login_sessions"))
        assert response.status_code == 200

    def test_audit(self):
        response = self.client.get(url_for("admin.audit"))
        assert response.status_code == 200

    def test_settings(self):
        response = self.client.get(url_for("admin.settings"))
        assert response.status_code == 200



# coding=utf-8

from __future__ import absolute_import, division, print_function, \
    unicode_literals

import pytest
import sqlalchemy as sa
from flask.signals import request_started
from wtforms import Form, IntegerField, StringField

# Import for side-effects (monkey-patch)
# noinspection PyUnresolvedReferences
import abilian.web.forms  # noqa
from abilian.core.entities import Entity
from abilian.testing import BaseTestCase
from abilian.web.views import default_view

from ..forms.widgets import EmailWidget, MainTableView, Panel, Row, \
    SingleView, linkify_url, text2html

skip = pytest.mark.skip


class WidgetTestModel(Entity):
    """Mock model."""
    __tablename__ = 'widget_test_model'
    id = sa.Column(sa.Integer, primary_key=True)
    price = sa.Column(sa.Integer)
    email = sa.Column(sa.Text)

    def __init__(self, *args, **kw):
        Entity.__init__(self, *args, **kw)
        self._display_value_called = False

    def display_value(self, attr):
        self._display_value_called = True
        return getattr(self, attr)


class DummyForm(Form):
    name = StringField('Nom du véhicule')
    price = IntegerField("Prix du véhicule")
    email = StringField('email', view_widget=EmailWidget())


class TableViewTestCase(BaseTestCase):

    def test_table_view(self):

        @default_view(self.app, WidgetTestModel)
        @self.app.route('/dummy_view/<object_id>')
        def dummy_view(object_id):
            pass

        with self.app.test_request_context():
            request_started.send(self.app)  # needed for deferJS tag
            columns = ['name', 'price']
            view = MainTableView(columns)

            model1 = WidgetTestModel(id=1, name="Renault Megane", price=10000)
            model2 = WidgetTestModel(id=2, name="Peugeot 308", price=12000)

            models = [model1, model2]

            res = view.render(models)

            assert model1._display_value_called
            assert model2._display_value_called
            assert "Renault Megane" in res
            assert "10000" in res


class ModelViewTestCase(BaseTestCase):
    # Hack to silence test harness bug
    __name__ = "ModelView test case"

    def test_single_view(self):
        with self.app.test_request_context():
            panels = [Panel('main', Row('name'), Row('price'), Row('email'))]
            view = SingleView(
                DummyForm, *panels, view=dict(can_edit=False, can_delete=False)
            )
            model = WidgetTestModel(
                name="Renault Megane",
                price=10000,
                email="joe@example.com",
            )
            form = DummyForm(obj=model)
            res = view.render(model, form)

            assert "Renault Megane" in res
            assert "10000" in res
            # 'mailto:' is created by EmailWidget
            assert "mailto:joe@example.com" in res

    @skip
    def test_edit_view(self):
        with self.app.test_request_context():
            panels = [Panel('main', Row('name'), Row('price'))]
            view = SingleView(DummyForm, *panels)
            model = WidgetTestModel(name="Renault Megane", price=10000)
            form = DummyForm(obj=model)
            res = view.render_form(form)

            assert model._display_value_called
            assert "Renault Megane" in res
            assert "10000" in res


EXPECTED = (
    '<a href="http://example.com">example.com</a>'
    '&nbsp;<i class="fa fa-external-link"></i>'
)


def test_http():
    value = "http://example.com"
    result = linkify_url(value)
    assert result == EXPECTED


def test_no_http():
    value = "example.com"
    result = linkify_url(value)
    assert result == EXPECTED


def test1():
    result = text2html("a")
    assert result == "a"


def test2():
    result = text2html("a\nb")
    assert str(result) == "<p>a</p>\n<p>b</p>"


def test3():
    result = text2html("a\n\nb")
    assert str(result) == "<p>a</p>\n<p>b</p>"


def test4():
    result = text2html("a\n<a>toto</a>")
    assert str(result) == "<p>a</p>\n<p>&lt;a&gt;toto&lt;/a&gt;</p>"



# coding=utf-8
"""Admin panel for tags."""
from __future__ import absolute_import, print_function, unicode_literals

import logging
from typing import List

import sqlalchemy as sa
import sqlalchemy.orm
import sqlalchemy.sql.functions as func
from flask import current_app, flash, redirect, render_template, request

from abilian.core.entities import Entity
from abilian.core.models.tag import Tag, entity_tag_tbl
from abilian.i18n import _, _l, _n
from abilian.services import get_service
from abilian.services.indexing.service import index_update
from abilian.web import url_for
from abilian.web.admin import AdminPanel
from abilian.web.views import ObjectEdit
from abilian.web.views.base import View

from .forms import TagForm

logger = logging.getLogger(__name__)

_OBJ_COUNT = func.count(entity_tag_tbl.c.entity_id) \
    .label('obj_count')


def get_entities_for_reindex(tags):
    """Collect entities for theses tags."""
    if isinstance(tags, Tag):
        tags = (tags,)

    session = current_app.db.session()
    indexing = get_service('indexing')
    tbl = Entity.__table__
    tag_ids = [t.id for t in tags]
    query = sa.sql \
        .select([tbl.c.entity_type, tbl.c.id]) \
        .select_from(tbl.join(
            entity_tag_tbl,
            entity_tag_tbl.c.entity_id == tbl.c.id,
        )) \
        .where(entity_tag_tbl.c.tag_id.in_(tag_ids))

    entities = set()

    with session.no_autoflush:
        for entity_type, entity_id in session.execute(query):
            if entity_type not in indexing.adapted:
                logger.debug('%r is not indexed, skipping', entity_type)

            item = ('changed', entity_type, entity_id, ())
            entities.add(item)

    return entities


def schedule_entities_reindex(entities):
    """
    :param entities: as returned by :func:`get_entities_for_reindex`
    """
    entities = [(e[0], e[1], e[2], dict(e[3])) for e in entities]
    return index_update.apply_async(
        kwargs=dict(
            index='default',
            items=entities,
        ),
    )


class NSView(View):
    """View a Namespace."""

    def __init__(self, view_endpoint, *args, **kwargs):
        super(NSView, self).__init__(*args, **kwargs)
        self.__selected_tags = None
        self.view_endpoint = view_endpoint

    def prepare_args(self, args, kwargs):
        self.ns = kwargs.get('ns')
        self.form_errors = {}
        return args, kwargs

    def get(self, ns):
        tags = Tag.query \
            .filter(Tag.ns == ns) \
            .outerjoin(entity_tag_tbl, entity_tag_tbl.c.tag_id == Tag.id) \
            .add_column(_OBJ_COUNT) \
            .group_by(Tag) \
            .order_by(sa.sql.func.lower(Tag.label))

        # get a list of rows instead of (Tag, count) tuples
        tags = list(tags.session.execute(tags))
        return render_template(
            'admin/tags_ns.html',
            ns=ns,
            tags=tags,
            errors=self.form_errors,
            merge_to=request.form.get(
                'merge_to', default='__None__', type=int),
            selected_tags={t.id
                           for t in self._get_selected_tags()},
        )

    def redirect_to_view(self):
        return redirect(url_for('.tags_ns', ns=self.ns))

    def post(self, ns):
        data = request.form
        action = data.get('__action')

        if action == 'delete':
            return self.do_delete()
        elif action == 'merge':
            return self.do_merge()

        else:
            flash(_('Unknown action'))
            self.get(self.ns)

    def _get_selected_tags(self):
        # type: () -> List[Tag]

        if self.__selected_tags is None:
            tag_ids = request.form.getlist('selected', type=int)
            if not tag_ids:
                self.__selected_tags = []
            else:
                self.__selected_tags = Tag.query \
                    .filter(Tag.ns == self.ns, Tag.id.in_(tag_ids)) \
                    .all()

        return self.__selected_tags

    def do_delete(self):
        data = request.form
        confirm = data.get('confirm_delete', False, type=bool)

        if not confirm:
            flash(_('Please fix the error(s) below'), 'error')
            self.form_errors['confirm_delete'] = _(
                'Must be checked to ensure you '
                'intent to delete these tags',
            )
            return self.get(self.ns)

        session = current_app.db.session()
        tags = self._get_selected_tags()

        if not tags:
            flash(_('No action performed: no tags selected'), 'warning')
            return self.redirect_to_view()

        count = len(tags)
        entities_to_reindex = get_entities_for_reindex(tags)
        success_message = _n(
            '%(tag)s deleted',
            '%(num)d tags deleted:\n%(tags)s',
            count,
            tag=tags[0].label,
            tags=', '.join(t.label for t in tags),
        )
        for tag in tags:
            session.delete(tag)
        session.commit()
        flash(success_message)
        schedule_entities_reindex(entities_to_reindex)
        return self.redirect_to_view()

    def do_merge(self):
        target_id = request.form.get('merge_to', type=int)

        if not target_id:
            flash(_('You must select a target tag to merge to'), 'error')
            return self.get(self.ns)

        target = Tag.query \
            .filter(Tag.ns == self.ns, Tag.id == target_id) \
            .scalar()

        if not target:
            flash(_('Target tag not found, no action performed'), 'error')
            return self.get(self.ns)

        merge_from = set(self._get_selected_tags())

        if target in merge_from:
            merge_from.remove(target)

        if not merge_from:
            flash(_('No tag selected for merging'), 'warning')
            return self.get(self.ns)

        session = current_app.db.session()
        merge_from_ids = [t.id for t in merge_from]
        tbl = entity_tag_tbl
        entities_to_reindex = get_entities_for_reindex(merge_from)
        already_tagged = sa.sql \
            .select([tbl.c.entity_id]) \
            .where(tbl.c.tag_id == target.id)
        del_dup = tbl.delete() \
            .where(sa.sql.and_(
                tbl.c.tag_id.in_(merge_from_ids),
                tbl.c.entity_id.in_(already_tagged),
            ))
        session.execute(del_dup)
        update = tbl.update() \
            .where(tbl.c.tag_id.in_(merge_from_ids)) \
            .values(tag_id=target.id)
        session.execute(update)
        for merged in merge_from:
            session.delete(merged)
        session.commit()
        schedule_entities_reindex(entities_to_reindex)
        return self.redirect_to_view()


class BaseTagView(object):
    """Mixin for tag views."""
    Model = Tag
    Form = TagForm

    def __init__(self, *args, **kwargs):
        super(BaseTagView, self).__init__(*args, **kwargs)
        self.extension = current_app.extensions['tags']

    def prepare_args(self, args, kwargs):
        self.ns = kwargs.pop('ns')
        return super(BaseTagView, self).prepare_args(args, kwargs)

    def view_url(self):
        return url_for('.tags_ns', ns=self.ns)

    index_url = view_url


class TagEdit(BaseTagView, ObjectEdit):
    _message_success = _l('Tag edited')
    has_changes = False
    _entities_to_reindex = []

    def after_populate_obj(self):
        session = sa.orm.object_session(self.obj)
        self.has_changes = self.obj in (session.dirty | session.deleted)
        if self.has_changes:
            # since the tag may be in pending-delete, we must collect them before
            # flush
            self._entities_to_reindex = get_entities_for_reindex(self.obj)

    def commit_success(self):
        if not (self.has_changes and self._entities_to_reindex):
            return

        schedule_entities_reindex(self._entities_to_reindex)


class TagPanel(AdminPanel):
    """Tags administration."""
    id = 'tags'
    label = _l('Tags')
    icon = 'tags'

    def get(self):
        obj_count = sa.sql \
            .select([
                Tag.ns,
                func.count(entity_tag_tbl.c.entity_id).label('obj_count'),
            ]) \
            .select_from(Tag.__table__.join(entity_tag_tbl)) \
            .group_by(Tag.ns) \
            .alias()

        ns_query = sa.sql \
            .select(
                [
                    Tag.ns,
                    func.count(Tag.id).label('tag_count'),
                    obj_count.c.obj_count,
                ],
                from_obj=[
                    Tag.__table__.outerjoin(
                        obj_count,
                        Tag.ns == obj_count.c.ns,
                    ),
                ],
            ) \
            .group_by(Tag.ns, obj_count.c.obj_count) \
            .order_by(Tag.ns)

        session = current_app.db.session()
        namespaces = session.execute(ns_query)

        return render_template('admin/tags.html', namespaces=namespaces)

    def install_additional_rules(self, add_url_rule):
        panel_endpoint = '.' + self.id
        ns_base = '/<string:ns>/'
        add_url_rule(
            ns_base,
            endpoint='ns',
            view_func=NSView.as_view('ns', view_endpoint=panel_endpoint),
        )

        tag_base = ns_base + '<int:object_id>/'
        add_url_rule(
            tag_base,
            endpoint='tag_edit',
            view_func=TagEdit.as_view(
                'tag_edit', view_endpoint=panel_endpoint),
        )

        add_url_rule(
            tag_base + 'delete',
            endpoint='tag_delete',
            view_func=TagEdit.as_view(
                'tag_delete',
                view_endpoint=panel_endpoint,
            ),
        )



# coding=utf-8
""""""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import time
from collections import deque

import sqlalchemy as sa
import whoosh
import whoosh.index
from flask import current_app
from six import string_types
from sqlalchemy.orm.session import Session
from tqdm import tqdm
from whoosh.writing import CLEAR, AsyncWriter

from abilian.services import get_service

from .base import manager

STOP = object()
COMMIT = object()


@manager.command
def reindex(clear=False, progressive=False, batch_size=None):
    """Reindex all content; optionally clear index before.

    All is done in asingle transaction by default.

    :param clear: clear index content.
    :param progressive: don't run in a single transaction.
    :param batch_size: number of documents to process before writing to the
                     index. Unused in single transaction mode. If `None` then
                     all documents of same content type are written at once.
    """
    index_service = get_service('indexing')
    adapted = index_service.adapted
    index = index_service.app_state.indexes['default']
    session = Session(
        bind=current_app.db.session.get_bind(None, None),
        autocommit=True,
    )

    session._model_changes = {}  # please flask-sqlalchemy <= 1.0
    indexed = set()
    cleared = set()
    if batch_size is not None:
        batch_size = int(batch_size)

    strategy = progressive_mode if progressive else single_transaction
    strategy = strategy(
        index,
        clear=clear,
        progressive=progressive,
        batch_size=batch_size,
    )
    next(strategy)  # starts generator

    for cls in sorted(
            index_service.app_state.indexed_classes,
            key=lambda c: c.__name__,
    ):
        current_object_type = cls._object_type()

        if not clear and current_object_type not in cleared:
            strategy.send(current_object_type)
            cleared.add(current_object_type)

        adapter = adapted.get(current_object_type)

        if not adapter or not adapter.indexable:
            continue

        name = cls.__name__

        with session.begin():
            query = session.query(cls).options(sa.orm.lazyload('*'))
            try:
                count = query.count()
            except Exception as e:
                current_app.logger.error(
                    "Indexing error on class {}: {}".format(name, repr(e)),
                )
                continue

            print("*" * 79)
            print("{}".format(name))
            if count == 0:
                print("*" * 79)
                print("{}".format(name))
                continue

            print("*" * 79)
            print("{}".format(name))
            count_current = 0
            with tqdm(total=count) as bar:
                for obj in query.yield_per(1000):
                    if obj.object_type != current_object_type:
                        # may happen if obj is a subclass and its parent class is also
                        # indexable
                        bar.update(1)
                        continue

                    object_key = obj.object_key

                    if object_key in indexed:
                        bar.update(1)
                        continue
                    document = index_service.get_document(obj, adapter)
                    strategy.send(document)
                    indexed.add(object_key)

                    if batch_size is not None and (
                            count_current % batch_size
                    ) == 0:
                        bar.update(1)
                        strategy.send(COMMIT)

                    bar.update(1)

            if batch_size is None:
                strategy.send(COMMIT)

        strategy.send(COMMIT)

    try:
        strategy.send(STOP)
    except StopIteration:
        pass

    try:
        strategy.close()
    except StopIteration:
        pass


# indexing strategies
def single_transaction(index, clear, **kwargs):
    with AsyncWriter(index) as writer:
        if clear:
            print("*" * 80)
            print("WILL CLEAR INDEX BEFORE REINDEXING")
            print("*" * 80)
            writer.writer.mergetype = CLEAR

        doc = yield True
        while doc is not STOP:
            if doc is COMMIT:
                doc = yield True
                continue
            if isinstance(doc, string_types):
                writer.delete_by_term('object_type', doc)
            else:
                writer.add_document(**doc)
            doc = yield True

        print("Writing Index...", end=' ')

    print("Done.")


def _get_writer(index):
    writer = None
    while writer is None:
        try:
            writer = index.writer()
        except whoosh.index.LockError:
            time.sleep(0.25)

    return writer


def progressive_mode(index, clear, batch_size, **kwargs):

    if clear:
        writer = _get_writer(index)
        print("*" * 80)
        print("CLEAR INDEX BEFORE REINDEXING")
        print("*" * 80)
        writer.mergetype = CLEAR
        writer.commit()
        del writer

    queue = deque()
    doc = yield True
    while doc is not STOP:
        if doc is COMMIT:
            writer = _get_writer(index)
            while queue:
                doc = queue.pop()
                if isinstance(doc, string_types):
                    writer.delete_by_term('object_type', doc)
                else:
                    writer.add_document(**doc)
            writer.commit()
            del writer
        else:
            queue.append(doc)

        doc = yield True



# coding=utf-8
""""""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from flask import current_app
from flask_login import current_user

from abilian.core.models import attachment as attachments
from abilian.services.security import READ, WRITE, security
from abilian.web import url_for

from .forms import AttachmentForm
from .views import bp as blueprint
from .views import UPLOAD_BUTTON

_MANAGER_ATTR = '__attachments_manager__'


class AttachmentExtension(object):
    """API for comments, installed as an application extension.

    It is also available in templates as `attachments`.
    """

    def __init__(self, app):
        app.extensions['attachments'] = self
        app.add_template_global(self, 'attachments')
        app.register_blueprint(blueprint)

    def manager(self, obj):
        """Returns the :class:`AttachmentsManager` instance for this object."""
        manager = getattr(obj, _MANAGER_ATTR, None)
        if manager is None:
            manager = AttachmentsManager()
            setattr(obj.__class__, _MANAGER_ATTR, manager)

        return manager

    def is_support_attachments(self, obj):
        return self.manager(obj).is_support_attachments(obj)

    def for_entity(self, obj, check_support_attachments=False):
        return self.manager(obj).for_entity(
            obj,
            check_support_attachments=check_support_attachments,
        )

    def has_attachments(self, obj):
        return self.manager(obj).has_attachments(obj)

    def count(self, obj):
        return self.manager(obj).count(obj)

    def get_form_context(self, obj):
        """Return a dict: form instance, action button, submit url...

        Used by macro m_attachment_form(entity)
        """
        return self.manager(obj).get_form_context(obj)


_DEFAULT_TEMPLATE = 'macros/attachment_default.html'


class AttachmentsManager(object):
    """Allow customization of attachments form, display macros, etc.

    can be used as class decorator
    """
    Form = AttachmentForm
    macros_template = 'macros/attachment.html'

    def __init__(
            self,
            Form=AttachmentForm,
            macros_template='macros/attachment_default.html',
    ):
        self.Form = Form
        self.macros_template = macros_template

    def __call__(self, Model):
        setattr(Model, _MANAGER_ATTR, self)
        return Model

    @property
    def macros(self):
        default_template = current_app.jinja_env.get_template(
            _DEFAULT_TEMPLATE)
        template = current_app.jinja_env.get_template(self.macros_template)
        default = default_template.module
        m = template.module
        return dict(
            m_attachments=getattr(m, 'm_attachments', default.m_attachments),
            m_attachment=getattr(m, 'm_attachment', default.m_attachment),
            m_attachment_form=getattr(
                m,
                'm_attachment_form',
                default.m_attachment_form,
            ),
        )

    def is_support_attachments(self, obj):
        return attachments.is_support_attachments(obj)

    def for_entity(self, obj, check_support_attachments=False):
        return attachments.for_entity(
            obj,
            check_support_attachments=check_support_attachments,
        )

    def has_attachments(self, obj):
        return bool(self.for_entity(obj, check_support_attachments=True))

    def count(self, obj):
        return len(self.for_entity(obj, check_support_attachments=True))

    def get_form_context(self, obj):
        """Return a dict: form instance, action button, submit url...

        Used by macro m_attachment_form(entity)
        """
        return {
            'url': url_for('attachments.create', entity_id=obj.id),
            'form': self.Form(),
            'buttons': [UPLOAD_BUTTON],
        }

    #
    # current user capabilities
    #
    def can_view(self, entity):
        """True if user can view attachments on entity."""
        return security.has_permission(current_user, READ, obj=entity)

    def can_edit(self, entity):
        """True if user can edit attachments on entity."""
        return security.has_permission(current_user, WRITE, obj=entity)

    def can_create(self, entity):
        """True if user can add attachments."""
        return security.has_permission(current_user, WRITE, obj=entity)

    def can_delete(self, entity):
        """True if user can delete attachments."""
        return security.has_permission(current_user, WRITE, obj=entity)



# coding=utf-8
"""
Base TestCase for integration tests.
"""
from __future__ import absolute_import, print_function, unicode_literals

import os

from abilian.testing import BaseTestCase

from .config import TestConfig

BASEDIR = os.path.dirname(__file__)


class IntegrationTestCase(BaseTestCase):
    config_class = TestConfig
    no_login = False

    def assert_302(self, response):
        assert response.status_core == 302

    def assert_204(self, response):
        assert response.status_core == 204



# coding=utf-8
"""Useful decorators for web views."""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import warnings
from functools import wraps


# Copy/pasted from:
# https://wiki.python.org/moin/PythonDecoratorLibrary#Generating_Deprecation_Warnings
def deprecated(func):
    """This decorator can be used to mark functions as deprecated.

    It will result in a warning being emitted when the function is used.
    """

    @wraps(func)
    def new_func(*args, **kwargs):
        warnings.warn_explicit(
            "Call to deprecated function {}.".format(func.__name__),
            category=DeprecationWarning,
            filename=func.__code__.co_filename,
            lineno=func.__code__.co_firstlineno + 1,
        )
        return func(*args, **kwargs)

    return new_func



# coding=utf-8
""""""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import pytest

from abilian.testing import BaseTestCase

from .models import EmptyValue, Setting


def test_type_set():
    s = Setting()
    # registered base type: no failure
    s.type = 'int'
    s.type = 'bool'
    s.type = 'json'
    s.type = 'string'

    with pytest.raises(ValueError):
        s.type = 'dummy type name'


def test_int():
    s = Setting(key='key', type='int')
    s.value = 42
    assert s._value in ('42', b'42')

    s._value = '24'
    assert s.value == 24


def test_bool():
    s = Setting(key='key', type='bool')
    s.value = True
    assert s._value == 'true'

    s.value = False
    assert s._value == 'false'

    s._value = 'true'
    assert s.value is True

    s._value = 'false'
    assert s.value is False


def test_string():
    s = Setting(key='key', type='string')
    s.value = 'ascii'
    assert s._value == b'ascii'

    s.value = 'bel été'
    assert s._value == b'bel \xc3\xa9t\xc3\xa9'

    s._value = b'd\xc3\xa9code'
    assert s.value == 'décode'


def test_json():
    s = Setting(key='key', type='json')
    s.value = [1, 2, 'été', {1: '1', 2: '2'}]
    assert s._value == '[1, 2, "\\u00e9t\\u00e9", {"1": "1", "2": "2"}]'

    s.value = None
    assert s._value == 'null'


def test_empty_value():
    s = Setting(key='key', type='json')
    s._value = None
    assert s.value == EmptyValue


class SettingsServiceTestCase(BaseTestCase):

    def setUp(self):
        BaseTestCase.setUp(self)
        self.service = self.app.services.get('settings')

    def test_service_facade(self):
        svc = self.service
        svc.set('key_1', 42, 'int')
        self.db.session.flush()
        assert svc.get('key_1') == 42

        # new key with no type: raise error:
        with pytest.raises(ValueError):
            svc.set('key_err', 42)

        # key already with type_, this should not raise an error
        svc.set('key_1', 24)
        self.db.session.flush()
        assert svc.get('key_1') == 24

        svc.delete('key_1')
        self.db.session.flush()
        with pytest.raises(KeyError):
            svc.get('key_1')

        # delete: silent by default
        svc.delete('non_existent')

        # delete: non-silent
        with pytest.raises(KeyError):
            svc.delete('non_existent', silent=False)

        # tricky use case: ask key delete, set it later, then flush
        svc.set('key_1', 42, 'int')
        self.db.session.flush()
        svc.delete('key_1')
        svc.set('key_1', 1)
        self.db.session.flush()
        assert svc.get('key_1') == 1

        # list keys
        svc.set('key_2', 2, 'int')
        svc.set('other', 'azerty', 'string')
        self.db.session.flush()
        assert sorted(svc.keys()) == ['key_1', 'key_2', 'other']
        assert sorted(svc.keys(prefix='key_')) == ['key_1', 'key_2']

        # as dict
        assert svc.as_dict() == {'other': 'azerty', 'key_1': 1, 'key_2': 2}
        assert svc.as_dict(prefix='key_') == {'key_1': 1, 'key_2': 2}

    def test_namespace(self):
        svc = self.service
        ns = svc.namespace('test')
        ns.set('1', 42, 'int')
        self.db.session.flush()
        assert ns.get('1') == 42
        assert svc.get('test:1') == 42

        ns.set('sub:2', 2, 'int')
        svc.set('other', 'not in NS', 'string')
        self.db.session.flush()
        assert sorted(ns.keys()) == ['1', 'sub:2']
        assert sorted(svc.keys()) == ['other', 'test:1', 'test:sub:2']

        # sub namespace test:sub:
        sub = ns.namespace('sub')
        assert sub.keys() == ['2']
        assert sub.get('2') == 2

        sub.set('1', 1, 'int')
        self.db.session.flush()
        assert sub.get('1') == 1
        assert ns.get('1') == 42
        assert sorted(
            svc.keys(),
        ) == ['other', 'test:1', 'test:sub:1', 'test:sub:2']

        # as dict
        assert sub.as_dict() == {'1': 1, '2': 2}
        assert ns.as_dict(prefix='sub:') == {'sub:1': 1, 'sub:2': 2}
        assert ns.as_dict() == {'1': 42, 'sub:1': 1, 'sub:2': 2}
        assert svc.as_dict() == {
            'other': 'not in NS',
            'test:1': 42,
            'test:sub:1': 1,
            'test:sub:2': 2,
        }

        # deletion
        sub.delete('1')
        sub.delete('2')
        self.db.session.flush()
        assert sub.keys() == []
        assert ns.keys() == ['1']
        assert sorted(svc.keys()) == ['other', 'test:1']



# coding=utf-8
"""Conversion service.

Hardcoded to manage only conversion to PDF, to text and to image series.

Includes result caching (on filesystem).

Assumes poppler-utils and LibreOffice are installed.

TODO: rename Converter into ConversionService ?
"""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import hashlib
import logging
import shutil
import subprocess
from io import BytesIO
from pathlib import Path

from PIL import Image
from PIL.ExifTags import TAGS
from six import string_types, text_type

from abilian.services.conversion.util import make_temp_file

logger = logging.getLogger(__name__)

TMP_DIR = "tmp"
CACHE_DIR = "cache"


class ConversionError(Exception):
    pass


class HandlerNotFound(ConversionError):
    pass


class Cache(object):

    CACHE_DIR = None

    def _path(self, key):
        """File path for `key`:"""
        return self.CACHE_DIR / "{}.blob".format(key)

    def __contains__(self, key):
        return self._path(key).exists()

    def get(self, key):
        if key in self:
            value = self._path(key).open('rb').read()
            if key.startswith("txt:"):
                value = text_type(value, encoding="utf8")
            return value
        else:
            return None

    __getitem__ = get

    def set(self, key, value):
        # if not os.path.exists(self.CACHE_DIR):
        #   os.mkdir(CACHE_DIR)
        fd = self._path(key).open("wb")
        if key.startswith("txt:"):
            fd.write(value.encode("utf8"))
        else:
            fd.write(value)
        fd.close()

    __setitem__ = set

    def clear(self):
        pass


class Converter(object):

    def __init__(self):
        self.handlers = []
        self.cache = Cache()

    def init_app(self, app):
        self.init_work_dirs(
            cache_dir=Path(app.instance_path, CACHE_DIR),
            tmp_dir=Path(app.instance_path, TMP_DIR),
        )

        app.extensions['conversion'] = self

        for handler in self.handlers:
            handler.init_app(app)

    def init_work_dirs(self, cache_dir, tmp_dir):
        self.TMP_DIR = Path(tmp_dir)
        self.CACHE_DIR = Path(cache_dir)
        self.cache.CACHE_DIR = self.CACHE_DIR

        if not self.TMP_DIR.exists():
            self.TMP_DIR.mkdir()
        if not self.CACHE_DIR.exists():
            self.CACHE_DIR.mkdir()

    def clear(self):
        self.cache.clear()
        for d in (self.TMP_DIR, self.CACHE_DIR):
            shutil.rmtree(bytes(d))
            d.mkdir()

    def register_handler(self, handler):
        self.handlers.append(handler)

    # TODO: refactor, pass a "File" or "Document" or "Blob" object
    def to_pdf(self, digest, blob, mime_type):
        cache_key = "pdf:" + digest
        pdf = self.cache.get(cache_key)
        if pdf:
            return pdf

        for handler in self.handlers:
            if handler.accept(mime_type, "application/pdf"):
                pdf = handler.convert(blob)
                self.cache[cache_key] = pdf
                return pdf
        raise HandlerNotFound(
            "No handler found to convert from %s to PDF" % mime_type,
        )

    def to_text(self, digest, blob, mime_type):
        """Convert a file to plain text.

        Useful for full-text indexing. Returns a Unicode string.
        """
        # Special case, for now (XXX).
        if mime_type.startswith("image/"):
            return ""

        cache_key = "txt:" + digest

        text = self.cache.get(cache_key)
        if text:
            return text

        # Direct conversion possible
        for handler in self.handlers:
            if handler.accept(mime_type, "text/plain"):
                text = handler.convert(blob)
                self.cache[cache_key] = text
                return text

        # Use PDF as a pivot format
        pdf = self.to_pdf(digest, blob, mime_type)
        for handler in self.handlers:
            if handler.accept("application/pdf", "text/plain"):
                text = handler.convert(pdf)
                self.cache[cache_key] = text
                return text

        raise HandlerNotFound(
            "No handler found to convert from %s to text" % mime_type,
        )

    def has_image(self, digest, mime_type, index, size=500):
        """Tell if there is a preview image."""
        cache_key = "img:%s:%s:%s" % (index, size, digest)
        return mime_type.startswith("image/") or cache_key in self.cache

    def get_image(self, digest, blob, mime_type, index, size=500):
        """Return an image for the given content, only if it already exists in
        the image cache."""
        # Special case, for now (XXX).
        if mime_type.startswith("image/"):
            return ""

        cache_key = "img:%s:%s:%s" % (index, size, digest)
        return self.cache.get(cache_key)

    def to_image(self, digest, blob, mime_type, index, size=500):
        """Convert a file to a list of images.

        Returns image at the given index.
        """
        # Special case, for now (XXX).
        if mime_type.startswith("image/"):
            return ""

        cache_key = "img:%s:%s:%s" % (index, size, digest)
        converted = self.cache.get(cache_key)
        if converted:
            return converted

        # Direct conversion possible
        for handler in self.handlers:
            if handler.accept(mime_type, "image/jpeg"):
                converted_images = handler.convert(blob, size=size)
                for i in range(0, len(converted_images)):
                    converted = converted_images[i]
                    self.cache["img:%s:%s:%s" % (i, size, digest)] = converted
                return converted_images[index]

        # Use PDF as a pivot format
        pdf = self.to_pdf(digest, blob, mime_type)
        for handler in self.handlers:
            if handler.accept("application/pdf", "image/jpeg"):
                converted_images = handler.convert(pdf, size=size)
                for i in range(0, len(converted_images)):
                    converted = converted_images[i]
                    self.cache["img:%s:%s:%s" % (i, size, digest)] = converted
                return converted_images[index]

        raise HandlerNotFound(
            "No handler found to convert from %s to image" % mime_type,
        )

    def get_metadata(self, digest, content, mime_type):
        """Get a dictionary representing the metadata embedded in the given
        content."""

        # XXX: ad-hoc for now, refactor later
        if mime_type.startswith("image/"):
            img = Image.open(BytesIO(content))
            ret = {}
            if not hasattr(img, '_getexif'):
                return {}
            info = img._getexif()
            if not info:
                return {}
            for tag, value in info.items():
                decoded = TAGS.get(tag, tag)
                ret["EXIF:" + str(decoded)] = value
            return ret
        else:
            if mime_type != "application/pdf":
                content = self.to_pdf(digest, content, mime_type)

            with make_temp_file(content) as in_fn:
                try:
                    output = subprocess.check_output(['pdfinfo', in_fn])
                except OSError:
                    logger.error(
                        "Conversion failed, probably pdfinfo is not installed",
                    )
                    raise

            ret = {}
            for line in output.split(b"\n"):
                if b":" in line:
                    key, value = line.strip().split(b":", 1)
                    key = text_type(key)
                    ret["PDF:" + key] = text_type(
                        value.strip(),
                        errors="replace",
                    )

            return ret

    @staticmethod
    def digest(blob):
        assert isinstance(blob, string_types)
        if isinstance(blob, str):
            digest = hashlib.md5(blob).hexdigest()
        else:
            digest = hashlib.md5(blob.encode("utf8")).hexdigest()
        return digest



# coding=utf-8
"""Abilian script commands to be used in a project.

See `Flask-Script documentation. <http://flask-script.readthedocs.org>`_
for full documentation.

Here is how a `manage.py` can include these commands::

     from flask_script import Manager
     from abilian.commands import setup_abilian_commands

     my_manager = Manager(app)
     setup_abilian_commands(my_manager)

You can also include abilian commands as sub commands::

     from abilian.commands import manager as abilian_manager
     my_manager.add_command('abilian', abilian_manager)

Extensions can add their own commands to :py:data:`~abilian.core.commands.manager`::

     from flask_script import Manager
     from abilian.commands import manager

     @manager.command
     def hello():
         print u"hello"

     # or install subcommands
     sub_manager = Manager(usage='Little extension')
     abilian_manager.add_command('special_commands', sub_manager)
"""
from __future__ import absolute_import, print_function, division

from flask_migrate import MigrateCommand
from flask_assets import ManageAssets

from .base import manager
from .config import manager as config_manager

# Additional commands
from . import indexing  # noqa

__all__ = ['manager', 'setup_abilian_commands']


def setup_abilian_commands(manager):
    """Register abilian commands on ``manager``.

    :param manager: ``flask_script.Manager`` instance to add commands onto

    Usage exemple::

        from flask_script import Manager
        from abilian.commands import setup_abilian_commands

        my_manager = Manager(app)
        setup_abilian_commands(my_manager)
    """
    abilian_manager = globals()['manager']
    manager._options.extend(abilian_manager._options)

    for name, command in abilian_manager._commands.items():
        manager.add_command(name, command)

    manager.add_command("assets", ManageAssets())  # flask-assets
    manager.add_command("config", config_manager)
    manager.add_command("migrate", MigrateCommand)
    return manager



# coding=utf-8
""""""
from __future__ import absolute_import, print_function, unicode_literals

import sqlalchemy as sa
from six import text_type

from abilian.core.models.tag import Tag, entity_tag_tbl
from abilian.i18n import _

from ..search.criterion import BaseCriterion
from .extension import ENTITY_DEFAULT_NS_ATTR


class TagCriterion(BaseCriterion):
    """Filter entities with selected tag(s)."""
    form_default_value = ''

    def __init__(self, *args, **kwargs):
        if len(args) == 0:
            kwargs.setdefault('name', 'tags')

        if len(args) < 2:
            kwargs.setdefault('label', _('Tags'))

        super(TagCriterion, self).__init__(*args, **kwargs)

    @BaseCriterion.model.setter
    def model(self, model):
        BaseCriterion.model.fset(self, model)
        self.ns = getattr(model, ENTITY_DEFAULT_NS_ATTR)

    @property
    def valid_tags(self):
        join_clause = entity_tag_tbl.join(
            self.model,
            self.model.id == entity_tag_tbl.c.entity_id,
        )
        model_tags = sa.sql.select(
            [entity_tag_tbl.c.tag_id],
            from_obj=join_clause,
        )
        return Tag.query.filter(
            Tag.ns == self.ns,
            Tag.id.in_(model_tags),
        )\
            .all()

    def get_request_values(self, request):
        tag_ids = []

        for val in request.values.getlist(self.name):
            try:
                tag_ids.append(int(val))
            except ValueError:
                pass

        valid_tags = set(self.valid_tags)
        tags = []
        for tid in tag_ids:
            t = Tag.query.get(tid)
            if t in valid_tags:
                tags.append(t)

        return tags

    def filter(self, query, module, request, searched_text, *args, **kwargs):

        tags = self.get_request_values(request)
        if not tags:
            return query

        cond = sa.sql.exists(
            sa.sql.select(
                [1],
                sa.sql.and_(
                    entity_tag_tbl.c.entity_id == self.model.id,
                    entity_tag_tbl.c.tag_id.in_(t.id for t in tags),
                ),
            ),
        )
        return query.filter(cond)

    @property
    def form_filter_type(self):
        return "select"

    @property
    def form_unset_value(self):
        return []

    @property
    def form_filter_args(self):
        # expected value: [list of selectable items, is multiple?]
        return [[(text_type(t.id), t.label) for t in self.valid_tags], True]



# coding=utf-8
""""""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from typing import Dict, List

import bleach
from wtforms.fields import BooleanField, IntegerField, StringField
from wtforms.widgets import HiddenInput

from abilian.i18n import _l
from abilian.web.forms import ModelForm
from abilian.web.forms.filters import strip
from abilian.web.forms.validators import required

ALLOWED_TAGS = ['b', 'i', 'del', 's', 'u', 'small', 'strong', 'em']
ALLOWED_ATTRIBUTES = {}  # type: Dict[str, List[str]]


class EditForm(ModelForm):
    label = StringField(
        _l('Label'),
        description=_l('allowed tags: %(tags)s', tags=', '.join(ALLOWED_TAGS)),
        filters=(strip,),
        validators=[required()],
    )
    default = BooleanField(_l('Default'), default=False)
    active = BooleanField(_l('Active'), default=True)

    def validate_label(self, field):
        field.data = bleach.clean(
            field.data,
            tags=ALLOWED_TAGS,
            attributes=ALLOWED_ATTRIBUTES,
            strip=True,
        )


class ListEditForm(EditForm):
    id = IntegerField(widget=HiddenInput())
    position = IntegerField(widget=HiddenInput())



# coding=utf-8
"""
This extensions makes intersphinx work with flask.ext extensions.

For example when actual module name is `flask_babel', documented
extension name is `flask.ext.babel`
"""
from __future__ import absolute_import


def missing_reference(app, env, node, contnode):
    target = node['reftarget']
    if not target.startswith('flask_'):
        return

    node['reftarget'] = 'flask.ext.' + target[6:]
    return


def setup(app):
    app.connect('missing-reference', missing_reference)



# coding=utf-8
"""Admin panel for vocabularies."""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from flask import current_app, g, redirect, render_template, request

from abilian.i18n import _, _l
from abilian.services import get_service
from abilian.web import url_for, views
from abilian.web.action import Glyphicon
from abilian.web.admin import AdminPanel
from abilian.web.nav import BreadcrumbItem

from .forms import EditForm

_MARKER = object()


class ViewBase(object):
    title = _l('Vocabulary entry')
    base_template = "admin/_base.html"
    Form = EditForm
    Model = None

    def prepare_args(self, args, kwargs):
        if self.Model is None:
            self.Model = kwargs.get('Model')
            if hasattr(self.Model, '__form__'):
                self.Form = getattr(self.Model, '__form__')
        return args, kwargs

    def index_url(self):
        return url_for('.vocabularies')


class Edit(ViewBase, views.ObjectEdit):

    def prepare_args(self, args, kwargs):
        args, kwargs = ViewBase.prepare_args(self, args, kwargs)
        return views.ObjectEdit.prepare_args(self, args, kwargs)

    def view_url(self):
        return url_for(
            '.vocabularies_model',
            group=self.Model.Meta.group or '_',
            Model=self.Model.Meta.name,
        )


class Create(views.ObjectCreate, Edit):

    def init_object(self, args, kwargs):
        args, kwargs = super(Create, self).init_object(args, kwargs)
        self.obj.active = True  # do this because default value seems ignored?
        return args, kwargs


class Delete(ViewBase, views.ObjectDelete):
    pass


class VocabularyPanel(AdminPanel):
    """Vocabularies administration."""
    id = 'vocabularies'
    label = _l('Vocabularies')
    icon = 'list'

    def voc_edit_url(self, item):
        return url_for(
            '.' + self.id + '_edit',
            group=item.Meta.group or '_',
            Model=item.Meta.name,
            object_id=item.id,
        )

    def get(self):
        svc = get_service('vocabularies')
        vocabularies = svc.grouped_vocabularies
        ctx = {
            'service': svc,
            'url_for_voc_edit': self.voc_edit_url,
            'icon_checked': Glyphicon('check'),
            'vocabularies': vocabularies,
        }
        return render_template('admin/vocabularies.html', **ctx)

    def post(self):
        data = request.form
        group = data.get('group', '').strip()
        Model = data.get('Model', '').strip()
        return_to = data.get('return_to')
        return_endpoint = '.vocabularies'
        return_args = {}

        if return_to not in (None, 'group', 'model'):
            return_to = None

        def do_return():
            return redirect(url_for(return_endpoint, **return_args))

        if not Model:
            return do_return()

        if not group or group == '_':
            # default group
            group = None

        svc = get_service('vocabularies')
        Model = svc.get_vocabulary(name=Model, group=group)
        if not Model:
            return do_return()

        if return_to is not None:
            return_endpoint += '_' + return_to

        if return_to == 'group':
            return_args['group'] = group or '_'
        elif return_to == 'model':
            return_args['group'] = Model.Meta.group or '_'
            return_args['Model'] = Model.Meta.name

        if 'up' in data:
            cmp_op = Model.position.__lt__
            cmp_order = Model.position.desc()
            object_id = int(data.get('up'))
        elif 'down' in data:
            cmp_op = Model.position.__gt__
            cmp_order = Model.position.asc()
            object_id = int(data.get('down'))
        else:
            return do_return()

        session = current_app.db.session()
        query = Model.query.with_lockmode('update')
        item = query.get(object_id)
        other = query \
            .filter(cmp_op(item.position)) \
            .order_by(cmp_order) \
            .first()

        if other is not None:
            # switch positions
            # we have to work around unique constraint on 'position', since we cannot
            # write new positions simultaneously
            # "-1" is added to avoid breaking when one position==0
            pos = other.position
            other.position = -item.position - 1
            item.position = -pos - 1
            session.flush()
            item.position = pos
            other.position = -other.position - 1
            session.commit()

        return do_return()

    def group_view(self, group):
        if group is None:
            group = ''
        svc = get_service('vocabularies')
        groups = svc.grouped_vocabularies
        vocabularies = groups.get(group)

        ctx = {
            'service': svc,
            'url_for_voc_edit': self.voc_edit_url,
            'icon_checked': Glyphicon('check'),
            'vocabularies': {
                group: vocabularies,
            },
            'edit_return_to': 'group',
        }
        return render_template('admin/vocabularies.html', **ctx)

    def model_view(self, Model, group=None):
        svc = get_service('vocabularies')

        ctx = {
            'service': svc,
            'url_for_voc_edit': self.voc_edit_url,
            'icon_checked': Glyphicon('check'),
            'vocabularies': {
                Model.Meta.group: [Model],
            },
            'edit_return_to': 'model',
        }
        return render_template('admin/vocabularies.html', **ctx)

    def install_additional_rules(self, add_url_rule):
        panel_endpoint = '.' + self.id
        group_base = '/<string:group>/'
        add_url_rule(group_base, endpoint='group', view_func=self.group_view)
        # models
        base = group_base + '<string:Model>/'
        add_url_rule(base, endpoint='model', view_func=self.model_view)

        edit_view = Edit.as_view('edit', view_endpoint=panel_endpoint)
        add_url_rule(base + '<int:object_id>', view_func=edit_view)
        add_url_rule(
            base + 'new',
            view_func=Create.as_view('new', view_endpoint=panel_endpoint),
        )

    def url_value_preprocess(self, endpoint, view_args):
        Model = view_args.pop('Model', None)
        group = view_args.pop('group', _MARKER)

        if group == '_':
            # "General" group
            group = None

        if group is not _MARKER:
            view_args['group'] = group

        if Model is not None:
            svc = get_service('vocabularies')
            Model = svc.get_vocabulary(name=Model, group=group)
            g.breadcrumb.append(BreadcrumbItem(
                label=Model.Meta.group if group else _('Global'),
                url=url_for('.vocabularies_group', group=group or '_'),
            ))
            g.breadcrumb.append(
                BreadcrumbItem(
                    label=Model.Meta.label,
                    url=url_for(
                        '.vocabularies_model',
                        group=group or '_',
                        Model=Model.Meta.name,
                    ),
                ),
            )
            view_args['Model'] = Model



# coding=utf-8
""""""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from flask import current_app
from werkzeug.exceptions import BadRequest

from abilian.core.entities import Entity
from abilian.core.models.tag import Tag
from abilian.i18n import _l
from abilian.web import url_for
from abilian.web.blueprints import Blueprint
from abilian.web.views import BaseObjectView, JSONView, ObjectCreate, \
    ObjectDelete, ObjectEdit

from .forms import TagForm

bp = Blueprint(
    'tags',
    __name__,
    url_prefix='/tags',
    template_folder='templates',
)


class BaseTagView(object):
    """Mixin for tag views."""
    Model = Tag
    Form = TagForm

    def __init__(self, *args, **kwargs):
        super(BaseTagView, self).__init__(*args, **kwargs)
        self.extension = current_app.extensions['tags']

    def view_url(self):
        return url_for()


class TagEdit(BaseTagView, ObjectEdit):
    _message_success = _l('Tag edited')


edit_view = TagEdit.as_view('edit')
bp.route('/manage/<int:object_id>/edit')(edit_view)


class TagCreate(BaseTagView, ObjectCreate):
    _message_success = _l('Tag created')


create_view = TagCreate.as_view('create')
bp.route('/manage/new')(create_view)


class TagDelete(BaseTagView, ObjectDelete):
    _message_success = _l('Tag deleted')


delete_view = TagDelete.as_view('delete')
bp.route('/manage/<int:object_id>/delete')(delete_view)

# Tags on entities
entity_bp = Blueprint('entity_tags', __name__, url_prefix='/tags/entity')


class BaseEntityTagView(BaseTagView):

    def init_object(self, args, kwargs):
        args, kwargs = super(BaseEntityTagView, self).init_object(args, kwargs)
        entity_id = kwargs.pop('entity_id', None)

        if entity_id is not None:
            self.entity = Entity.query.get(entity_id)

        if self.entity is None:
            raise BadRequest('No entity provided')

        return args, kwargs

    def view_url(self):
        return url_for(self.entity)

    def index_url(self):
        return self.view_url()


class EntityTagList(BaseEntityTagView, BaseObjectView, JSONView):

    def get(self, *args, **kwargs):
        return JSONView.get(self, *args, **kwargs)

    def data(self, *args, **kwargs):
        tags = sorted(self.extension.entity_tags(self.entity))
        return dict(result=tags)


entity_bp.route('/<int:object_id>/list')(EntityTagList.as_view('list'))


class EntityTagManage(BaseEntityTagView, ObjectEdit):
    methods = ['POST']

    # operation: add or remove
    mode = None

    def __init__(self, mode, *args, **kwargs):
        super(EntityTagManage, *args, **kwargs)
        assert mode in ('add', 'remove')
        self.mode = mode

    def form_valid(self, redirect_to=None):
        ns = self.form.ns.data
        label = self.form.label.data
        op = getattr(self.extension, self.mode)
        op(self.entity, ns=ns, label=label)


entity_bp.route('/<int:object_id>/add')(
    EntityTagManage.as_view(
        'add',
        mode='add',
    ),
)
entity_bp.route('/<int:object_id>/remove')(
    EntityTagManage.as_view(
        'remove',
        mode='remove',
    ),
)


class EntityTagEdit(ObjectEdit):
    Model = Entity

    def init_object(self, args, kwargs):
        args, kwargs = super(EntityTagEdit, self).init_object(args, kwargs)
        extension = current_app.extensions['tags']
        self.Form = extension.entity_tags_form(self.obj)
        return args, kwargs

    def view_url(self):
        return url_for(self.obj)

    def index_url(self):
        return self.view_url()


entity_bp.route('/<int:object_id>/edit')(EntityTagEdit.as_view('edit'))



# coding=utf-8
""""""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from typing import Optional

import sqlalchemy as sa
import sqlalchemy.orm
from flask import current_app, send_file
from werkzeug.exceptions import BadRequest
from werkzeug.utils import redirect

from abilian.core.entities import Entity
from abilian.core.models.attachment import Attachment, is_support_attachments
from abilian.i18n import _, _l
from abilian.web import nav, url_for
from abilian.web.action import ButtonAction, actions
from abilian.web.blueprints import Blueprint
from abilian.web.views import BaseObjectView, ObjectCreate, ObjectDelete, \
    ObjectEdit

from .forms import AttachmentForm

bp = Blueprint('attachments', __name__, url_prefix='/attachments')


def _default_attachment_view(obj, obj_type, obj_id, **kwargs):
    if not hasattr(obj, 'entity'):
        return url_for('attachments.entity', object_id=obj_id)
    entity = obj.entity
    return url_for(entity, _anchor='attachment-{}'.format(obj.id))


@bp.record_once
def register_default_view(state):
    state.app.default_view.register(Attachment, _default_attachment_view)


UPLOAD_BUTTON = ButtonAction(
    'form',
    'edit',
    btn_class='primary',
    title=_l('Send'),
)


class BaseAttachmentView(object):
    """Mixin for attachment views."""
    Model = Attachment
    Form = AttachmentForm

    #: owning entity
    entity = None  # type: Optional[Entity]

    def init_object(self, args, kwargs):
        args, kwargs = super(
            BaseAttachmentView, self).init_object(
            args, kwargs)
        entity_id = kwargs.pop('entity_id', None)

        if entity_id is not None:
            self.entity = Entity.query.get(entity_id)

        if self.entity is None:
            raise BadRequest('No entity provided')

        if not is_support_attachments(self.entity):
            raise BadRequest('This entity is doesn\'t support attachments')

        extension = current_app.extensions['attachments']
        self.Form = extension.manager(self.entity).Form
        actions.context['object'] = self.entity
        return args, kwargs

    def view_url(self):
        kw = {}
        if self.obj and self.obj.id:
            kw['_anchor'] = 'attachment-{}'.format(self.obj.id)
        return url_for(self.entity, **kw)

    def index_url(self):
        return self.view_url()

    @property
    def activity_target(self):
        return self.entity


class AttachmentDownload(BaseAttachmentView, BaseObjectView):

    def get(self):
        blob = self.obj.blob
        metadata = blob.meta
        filename = metadata.get('filename', self.obj.name)
        content_type = metadata.get('mimetype')
        stream = blob.file.open('rb')

        return send_file(
            stream,
            as_attachment=True,
            attachment_filename=filename,
            mimetype=content_type,
            cache_timeout=0,
            add_etags=False,
            conditional=False,
        )


download_view = AttachmentDownload.as_view('download')
bp.route('/<int:entity_id>/<int:object_id>/download')(download_view)


class AttachmentEdit(BaseAttachmentView, ObjectEdit):
    _message_success = _l('Attachment edited')


edit_view = AttachmentEdit.as_view('edit')
bp.route('/<int:entity_id>/<int:object_id>/edit')(edit_view)


class AttachmentCreateView(BaseAttachmentView, ObjectCreate):

    _message_success = _l('Attachment added')

    def init_object(self, args, kwargs):
        args, kwargs = super(AttachmentCreateView, self).init_object(
            args,
            kwargs,
        )

        self.obj.entity = self.entity
        session = sa.orm.object_session(self.entity)

        if session:
            sa.orm.session.make_transient(self.obj)

        return args, kwargs

    def breadcrumb(self):
        label = _('New attachment on "{title}"').format(title=self.entity.name)
        return nav.BreadcrumbItem(label=label)

    def get_form_buttons(self, *args, **kwargs):
        return [UPLOAD_BUTTON]


create_view = AttachmentCreateView.as_view('create')
bp.route('/<int:entity_id>/create')(create_view)


class AttachmentDelete(BaseAttachmentView, ObjectDelete):
    pass


delete_view = AttachmentDelete.as_view('delete')
bp.route('/<int:entity_id>/<int:object_id>/delete')(delete_view)


class AttachmentEntity(BaseObjectView):
    """Redirects to an attachment's entity view."""
    Model = Attachment

    def get(self):
        return redirect(url_for(self.obj))


entity_view = AttachmentEntity.as_view('entity')
bp.route('/<int:object_id>/entity')(entity_view)



# coding=utf-8
""""""
from __future__ import absolute_import, division, print_function

from .models import Vocabulary
from .service import get_vocabulary, vocabularies

__all__ = ['vocabularies', 'get_vocabulary', 'Vocabulary']



# coding=utf-8
""""""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from abilian.core.entities import Entity
from abilian.services import security
from abilian.testing import BaseTestCase as AbilianTestCase


class PermissionsTestCase(AbilianTestCase):

    def test_default_permissions(self):

        class MyRestrictedType(Entity):
            __default_permissions__ = {
                security.READ: {security.Anonymous},
                security.WRITE: {security.Owner},
                security.CREATE: {security.Writer},
                security.DELETE: {security.Owner},
            }

        assert isinstance(MyRestrictedType.__default_permissions__, frozenset)

        expected = frozenset({
            (security.READ, frozenset({security.Anonymous})),
            #
            (security.WRITE, frozenset({security.Owner})),
            #
            (security.CREATE, frozenset({security.Writer})),
            #
            (security.DELETE, frozenset({security.Owner})),
        })
        assert MyRestrictedType.__default_permissions__ == expected

        self.app.db.create_all()  # create missing 'mytype' table

        obj = MyRestrictedType(name='test object')
        self.session.add(obj)
        PA = security.PermissionAssignment
        query = self.session.query(PA.role) \
            .filter(PA.object == obj)

        assert query.filter(PA.permission == security.READ).all() \
            == [(security.Anonymous,)]

        assert query.filter(PA.permission == security.WRITE).all() \
            == [(security.Owner,)]

        assert query.filter(PA.permission == security.DELETE).all() \
            == [(security.Owner,)]

        # special case:
        assert query.filter(PA.permission == security.CREATE).all() \
            == []

        security_svc = self.app.services['security']
        permissions = security_svc.get_permissions_assignments(obj)
        assert permissions == {
            security.READ: {security.Anonymous},
            security.WRITE: {security.Owner},
            security.DELETE: {security.Owner},
        }



# coding=utf-8
""""""
from __future__ import absolute_import, division, print_function

from flask import render_template

from abilian.i18n import _, _l
from abilian.web.admin.panel import AdminPanel
from abilian.web.util import url_for

from . import views


class UsersPanel(AdminPanel):
    """User administration panel."""
    id = 'users'
    label = _l(u'Users')
    icon = 'user'

    def install_additional_rules(self, add_url_rule):
        add_url_rule(
            '/users',
            view_func=views.JsonUsersList.as_view('json_list'),
        )
        add_url_rule('/new', view_func=views.UserCreate.as_view('new'))
        add_url_rule(
            '/<int:user_id>',
            view_func=views.UserEdit.as_view('user'))

    def get(self):
        # FIXME: use widgets.AjaxMainTableView instead
        datatable_options = {
            'sDom': 'lfFritip',
            'aaSorting': [
                [1, u'asc'],
            ],
            'aoColumns': [
                dict(bSortable=False),
                dict(asSorting=['asc', 'desc']),
                dict(asSorting=['asc', 'desc']),
                dict(bSortable=False),
                dict(bSortable=False),
                dict(bSortable=False),
                dict(asSorting=['asc', 'desc']),
            ],
            'bFilter': True,
            'oLanguage': {
                'sSearch': _("Filter records:"),
                'sPrevious': _("Previous"),
                'sNext': _("Next"),
                'sInfo': _("Showing _START_ to _END_ of _TOTAL_ entries"),
                'sInfoFiltered': _("(filtered from _MAX_ total entries)"),
                'sAddAdvancedFilter': _("Add a filter"),
            },
            'bStateSave': False,
            'bPaginate': True,
            'sPaginationType': "bootstrap",
            'bLengthChange': False,
            'iDisplayLength': 30,
            'bProcessing': True,
            'bServerSide': True,
            'sAjaxSource': url_for('.users_json_list'),
        }

        return render_template(
            'admin/users.html',
            next=next,
            datatable_options=datatable_options,
        )



# coding=utf-8
""""""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import uuid
from pathlib import Path

import pytest

from abilian.testing import BaseTestCase

from . import repository, session_repository
from .service import RepositoryTransaction


class TestRepository(BaseTestCase):

    UUID_STR = '4f80f02f-52e3-4fe2-b9f2-2c3e99449ce9'
    UUID = uuid.UUID(UUID_STR)

    def test_rel_path(self):
        with pytest.raises(ValueError):
            repository.rel_path(self.UUID_STR)

        p = repository.rel_path(self.UUID)
        expected = Path('4f', '80', '4f80f02f-52e3-4fe2-b9f2-2c3e99449ce9')
        assert isinstance(p, Path)
        assert p == expected

    def test_abs_path(self):
        with pytest.raises(ValueError):
            repository.abs_path(self.UUID_STR)

        p = repository.abs_path(self.UUID)
        assert isinstance(p, Path)

        # FIXME: fails on a Mac due to symlinks /var@ -> /private/var
        # expected = Path(self.app.instance_path, 'data', 'files',
        #                 '4f', '80', '4f80f02f-52e3-4fe2-b9f2-2c3e99449ce9')
        # self.assertEquals(p, expected)

    def test_get(self):
        with pytest.raises(ValueError):
            repository.get(self.UUID_STR)

        with pytest.raises(ValueError):
            repository[self.UUID_STR]

        p = repository.abs_path(self.UUID)
        if not p.parent.exists():
            p.parent.mkdir(parents=True)
        p.open('bw').write(b'my file content')

        val = repository.get(self.UUID)
        assert val == p
        assert val.open('rb').read() == b'my file content'

        # non-existent
        u = uuid.UUID('bcdc32ac-498d-4544-9e7f-fb2c75097011')
        null = object()
        assert repository.get(u) is None
        assert repository.get(u, default=null) is null

        # __getitem__
        val = repository[self.UUID]
        assert val == p
        assert val.open('rb').read() == b'my file content'

        # __getitem__ non-existent
        with pytest.raises(KeyError):
            repository[u]

    def test_set(self):
        with pytest.raises(ValueError):
            repository.set(self.UUID_STR, '')
        with pytest.raises(ValueError):
            repository[self.UUID_STR] = ''

        p = repository.abs_path(self.UUID)
        repository.set(self.UUID, b'my file content')
        assert p.open('rb').read() == b'my file content'

        # __setitem__
        p.unlink()
        assert p.exists() == False

        repository[self.UUID] = b'my file content'
        assert p.open('rb').read() == b'my file content'
        # FIXME: test Unicode content

    def test_delete(self):
        with pytest.raises(ValueError):
            repository.delete(self.UUID_STR)
        with pytest.raises(ValueError):
            del repository[self.UUID_STR]

        p = repository.abs_path(self.UUID)
        p.parent.mkdir(parents=True)
        p.open('bw').write(b'my file content')
        repository.delete(self.UUID)
        assert not p.exists()
        assert p.parent.exists()

        # non-existent
        u = uuid.UUID('bcdc32ac-498d-4544-9e7f-fb2c75097011')
        with pytest.raises(KeyError):
            repository.delete(u)

        # __delitem__
        p.open('bw').write(b'my file content')
        assert p.exists()

        del repository[self.UUID]
        assert not p.exists()
        assert p.parent.exists()

        # __delitem__ non-existent
        with pytest.raises(KeyError):
            del repository[u]


class TestSessionRepository(BaseTestCase):

    UUID_STR = '4f80f02f-52e3-4fe2-b9f2-2c3e99449ce9'
    UUID = uuid.UUID(UUID_STR)

    def setUp(self):
        BaseTestCase.setUp(self)
        self.svc = session_repository

    def test_transaction_lifetime(self):
        state = self.svc.app_state
        root_transaction = state.get_transaction(self.session)
        assert isinstance(root_transaction, RepositoryTransaction)
        assert root_transaction._parent is None

        # create sub-transaction (db savepoint)
        self.session.begin(nested=True)
        transaction = state.get_transaction(self.session)
        assert isinstance(transaction, RepositoryTransaction)
        assert transaction._parent is root_transaction

        self.session.commit()
        transaction = state.get_transaction(self.session)
        assert transaction is root_transaction

        # create subtransaction (sqlalchemy)
        self.session.begin(subtransactions=True)
        transaction = state.get_transaction(self.session)
        assert isinstance(transaction, RepositoryTransaction)
        assert transaction._parent is root_transaction

        self.session.commit()
        transaction = state.get_transaction(self.session)
        assert transaction is root_transaction

    def test_accessors(self):
        session = self.session
        with pytest.raises(ValueError):
            self.svc.get(session, self.UUID_STR)
        with pytest.raises(ValueError):
            self.svc.set(session, self.UUID_STR, '')
        with pytest.raises(ValueError):
            self.svc.delete(session, self.UUID_STR)

        # non-existent
        u = uuid.UUID('bcdc32ac-498d-4544-9e7f-fb2c75097011')
        null = object()
        assert self.svc.get(session, u) is None
        assert self.svc.get(session, u, default=null) is null

        # set
        self.svc.set(session, self.UUID, b'my file content')
        assert self.svc.get(session, self.UUID).open('rb').read() == \
            b'my file content'
        assert repository.get(self.UUID) is None

        # delete
        self.svc.delete(session, self.UUID)
        assert self.svc.get(session, self.UUID) is None

        u = uuid.UUID('2532e752-611d-469c-999a-74f651757dff')
        repository.set(u, b'existing content')
        assert self.svc.get(session, u) is not None

        self.svc.delete(session, u)
        assert self.svc.get(session, u) is None
        assert repository.get(u) is not None

    def test_transaction(self):
        session = self.session
        repository.set(self.UUID, b'first draft')
        assert self.svc.get(session, self.UUID).open('rb').read() == \
            b'first draft'

        self.svc.set(session, self.UUID, b'new content')

        # test nested (savepoint)
        # delete content but rollback transaction
        db_tr = session.begin(nested=True)
        self.svc.delete(session, self.UUID)
        assert self.svc.get(session, self.UUID) is None

        db_tr.rollback()
        assert self.svc.get(session, self.UUID).open('rb').read() == \
            b'new content'

        # delete and commit
        with session.begin(nested=True):
            self.svc.delete(session, self.UUID)
            assert self.svc.get(session, self.UUID) is None

        assert self.svc.get(session, self.UUID) is None
        assert repository.get(self.UUID) is not None

        session.commit()
        assert repository.get(self.UUID) is None

        # delete: now test subtransactions (sqlalchemy)
        repository.set(self.UUID, b'first draft')
        db_tr = session.begin(subtransactions=True)
        self.svc.delete(session, self.UUID)
        assert self.svc.get(session, self.UUID) is None

        db_tr.rollback()
        assert self.svc.get(session, self.UUID).open('rb').read() == \
            b'first draft'

        session.rollback()

        with session.begin(subtransactions=True):
            self.svc.delete(session, self.UUID)
            assert self.svc.get(session, self.UUID) is None

        assert self.svc.get(session, self.UUID) is None
        assert repository.get(self.UUID) is not None

        session.commit()
        assert repository.get(self.UUID) is None

        # now test 'set'
        self.svc.set(session, self.UUID, b'new content')
        session.commit()
        assert repository.get(self.UUID) is not None

        # test "set" in two nested transactions. This tests a specific code branch,
        # when a subtransaction overwrite data set in parent transaction
        with session.begin(nested=True):
            self.svc.set(session, self.UUID, b'transaction 1')

            with session.begin(nested=True):
                self.svc.set(session, self.UUID, b'transaction 2')

            assert self.svc.get(session, self.UUID).open('rb').read() == \
                b'transaction 2'

    def test_transaction_path(self):
        """Test RepositoryTransaction create storage only when needed."""
        session = self.session
        state = self.svc.app_state
        root_transaction = state.get_transaction(self.session)

        assert not root_transaction.path.exists()

        with session.begin(subtransactions=True):
            transaction = state.get_transaction(self.session)
            assert not transaction.path.exists()

            self.svc.set(session, self.UUID, b'my file content')
            assert transaction.path.exists()

        assert root_transaction.path.exists()

        content = self.svc.get(session, self.UUID).open('rb').read()
        assert content == b'my file content'
        assert root_transaction.path.exists()



# coding=utf-8

from __future__ import absolute_import, print_function, unicode_literals

import pytest
from wtforms import Form, StringField
from wtforms.validators import ValidationError

from abilian.web.forms.validators import siret_validator


class DummyForm(Form):
    name = StringField('foo')
    siret = StringField('Siret')


class DummyField(object):

    def __init__(self, data, errors=(), raw_data=None):
        self.data = data
        self.errors = list(errors)
        self.raw_data = raw_data


def error_message(validator, form, field):
    try:
        validator(form, field)
    except ValidationError as e:
        return e.args[0]


@pytest.fixture()
def validator():
    yield siret_validator()


def test_siret_validator_valid_1(validator):
    # valid
    form = DummyForm(siret="54207855500514", name="foo")
    field = form.siret
    assert validator(form, field) is None


def test_siret_validator_valid_2(validator):
    # valid
    form = DummyForm(siret="54207855500514", name="foo")
    field = form.siret
    assert validator(form, field) is None


def test_siret_validator_invalid_luhn(validator):
    # invalid Luhn (changed the first digit)
    form = DummyForm(siret="64207855500514", name="foo")
    field = form.siret
    with pytest.raises(ValidationError):
        validator(form, field)


def test_siret_validator_invalid_2(validator):
    # invalid
    form = DummyForm(siret="WRONG542078555", name="foo")
    field = form.siret
    with pytest.raises(ValidationError):
        validator(form, field)


def test_siret_validator_too_short(validator):
    # too short
    form = DummyForm(siret="54207", name="foo")
    field = form.siret
    with pytest.raises(ValidationError):
        validator(form, field)


def test_siret_ok():
    siret = siret_validator()
    form = DummyForm()
    assert siret(form, DummyField('78913349300013')) is None
    assert siret(form, DummyField('MONACOCONFO001')) is None

    # test other geographical exceptions; WARNING! the siret format is probably
    # not right, but we had no example in the spec; only geo codes...
    assert siret(form, DummyField('MONACOCONFO458')) is None
    assert siret(form, DummyField('MONACOCONFO462')) is None
    assert siret(form, DummyField('MONACOCONFO496')) is None
    assert siret(form, DummyField('MONACOCONFO372')) is None


def test_siret_ko_special_siret(validator):
    form = DummyForm()

    field = DummyField('MONACOCONFO999')
    with pytest.raises(ValidationError):
        validator(form, field)
    assert error_message(validator, form, field) == \
        'SIRET looks like special SIRET but geographical code seems invalid (999)'


def test_siret_ko_invalid_length(validator):
    # invalid length
    form = DummyForm()
    field = DummyField('42342435')
    with pytest.raises(ValidationError):
        validator(form, field)
    assert error_message(validator, form, field) == \
        'SIRET must have exactly 14 characters (8)'


def test_siret_ko_invalid_luhn(validator):
    # invalid checksum
    form = DummyForm()
    field = DummyField('78913349300011')
    with pytest.raises(ValidationError):
        validator(form, field)
    assert error_message(validator, form, field) == \
        'SIRET number is invalid (length is ok: verify numbers)'



# coding=utf-8
""""""
from __future__ import absolute_import, division, print_function

from .criterion import TagCriterion
from .extension import TagsExtension

__all__ = ['TagsExtension', 'TagCriterion']


def register_plugin(app):
    TagsExtension(app=app)
